from lxml import html, etree
from parslepy import Parselet
from parslepy.funcs import xpathstrip, xpathtostring
from urlparse import urldefrag, urljoin
from cssselect import HTMLTranslator
import re


ns = etree.FunctionNamespace(None)
ns['strip'] = xpathstrip
ns['str'] = xpathtostring


def extract_urls(self, xpath=''):
    urlpattern = re.compile(
        r'^(?:http)s?://'  # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
        r'(?::\d+)?'  # optional port
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)
    if xpath and not xpath.endswith('/'):
        xpath += '/'
    return set(url.split('#')[0] for url in
               self.xpath(xpath + "descendant-or-self::a/@href")
               if urlpattern.match(url))


def extract_text(self):
    return "".join(i.strip() for i in self.itertext())


def extract(self, rules, strict=False):
    parselet = Parselet(rules, strict=strict)
    return parselet.extract(self)


def cssselect(self, expr):
    return self._css_translator.css_to_xpath(expr)


def css(self, expr):
    return self.xpath(self.cssselect(expr))


html.HtmlElement.extract_text = extract_text
html.HtmlElement._css_translator = HTMLTranslator()
html.HtmlElement.cssselect = cssselect
html.HtmlElement.css = css
html.HtmlElement.extract = extract
html.HtmlElement.extract_urls = extract_urls


def HtmlParser(response):
    """
    :param response:
    :type response: :class:`dragline.http.Response`

    This method takes response object as its argument and returns
    the lxml etree object.

    HtmlParser function returns a lxml object of type Element which got 2 potential methods.
    All the details of lxml object are discussed in section `lxml Element object`_.
    """
    element = html.fromstring(response.body, response.url)
    element.make_links_absolute()
    return element



from urllib import urlencode
import socket
from hashlib import sha1
import time
import httplib2
from defaultsettings import RequestSettings


class RequestError(Exception):

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class Request:

    """
    :param url: the URL of this request
    :type url: string
    :param method: the HTTP method of this request. Defaults to ``'GET'``.
    :type method: string
    :param headers: the headers of this request.
    :type headers: dict
    :param callback: name of the function to call after url is downloaded.
    :type callback: string
    :param meta:  A dict that contains arbitrary metadata for this request.
    :type meta: dict
    """

    settings = RequestSettings()
    method = "GET"
    form_data = None
    headers = {}
    callback = None
    meta = None
    retry = 0

    def __init__(self, url, method="GET", form_data=None, headers={}, callback=None, meta=None,):
        self.url = url
        if method:
            self.method = method
        if callback:
            self.callback = callback
        if meta:
            self.meta = meta
        if form_data:
            self.form_data = form_data
        if headers:
            self.headers = headers

    def _set_state(self, state):
        self.__dict__ = state

    def __str__(self):
        return self.get_unique_id(False)

    def usha1(self, x):
        """sha1 with unicode support"""
        if isinstance(x, unicode):
            return sha1(x.encode('utf-8')).hexdigest()
        else:
            return sha1(x).hexdigest()

    def send(self):
        """
        This function sends HTTP requests.

        :returns: response
        :rtype: :class:`dragline.http.Response`
        :raises: :exc:`dragline.http.RequestError`: when failed to fetch contents

        >>> req = Request("http://www.example.org")
        >>> response = req.send()
        >>> print response.headers['status']
        200

        """
        form_data = urlencode(self.form_data) if self.form_data else None
        try:
            time.sleep(Request.settings.DELAY)
            start = time.time()
            http = httplib2.Http(cache=self.settings.CACHE)
            req_headers = self.settings.HEADERS
            req_headers.update(self.headers)
            headers, content = http.request(
                self.url, self.method, form_data, req_headers)
            res = Response(self.url, content, headers, self.meta)
            end = time.time()
            self.updatedelay(end, start)
        except (httplib2.ServerNotFoundError, socket.timeout, socket.gaierror) as e:
            raise RequestError(e.message)
        return res

    def get_unique_id(self, hashing=True):
        request = self.method + ":" + self.url
        if self.form_data:
            request += ":" + urlencode(
                {i: j for i, j in sorted(self.form_data.items(),
                                         key=lambda t: t[0])})
        if hashing:
            return self.usha1(request)
        else:
            return request

    @classmethod
    def updatedelay(cls, end, start):
        delay = end - start
        cls.settings.DELAY = min(
            max(cls.settings.MIN_DELAY, delay,
                (cls.settings.DELAY + delay) / 2.0),
            cls.settings.MAX_DELAY)


class Response:

    """
    This function is used to create user defined 
    respones to test your spider and also in many 
    other cases.
    :param url: the URL of this response
    :type url: string

    :param headers: the headers of this response.
    :type headers: dict

    :param body: the response body.
    :type body: str

    :param meta: meta copied from request
    :type meta: dict

    """

    def __init__(self, url=None, body=None, headers=None, meta=None):
        self.url = url
        self.body = body
        self.headers = headers
        self.meta = meta



from gevent import monkey, spawn, joinall
monkey.patch_all()

import sys
import argparse
import os
import traceback
import logging
import logging.config
from defaultsettings import SpiderSettings
from crawl import Crawler


logging.config.dictConfig(SpiderSettings.LOGCONFIG)
logger = logging.getLogger("dragline")


def load_module(path, filename):
    try:
        sys.path.insert(0, path)
        module = __import__(filename)
        del sys.path[0]
        return module
    except Exception as e:
        logger.exception("Failed to load module %s" % filename)
        raise ImportError


def main(directory):
    settings_module = load_module(directory, "settings")
    spider_module = load_module(directory, "main")
    spider = getattr(spider_module, "Spider")
    Crawler.load_spider(spider, settings_module)
    crawlers = [Crawler() for i in xrange(5)]
    try:
        joinall([spawn(crawler.process_url) for crawler in crawlers])
    except KeyboardInterrupt:
        Crawler.crawl.clear(False)
    except:
        logger.exception("Unable to complete")
    else:
        Crawler.crawl.clear(True)
        logger.info("Crawling completed")


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument('spider', help='spider directory name')
    args = parser.parse_args()
    path = os.path.abspath(args.spider)
    main(path)

if __name__ == "__main__":
    run()



from string import Template
import os
import argparse
import dragline


def generate(spider_name):
    template_dir = os.path.join(dragline.__path__[0], "templates")
    main = open(os.path.join(template_dir, "main.tem"), "r").read()
    s = Template(main)
    main = s.substitute(spider_name=spider_name)
    settings = open(os.path.join(template_dir, "settings.tem"), "r").read()

    os.makedirs(spider_name)
    mainfile = open(spider_name + "/main.py", "w")
    mainfile.write(main)
    mainfile.close()
    settfile = open(spider_name + "/settings.py", "w")
    settfile.write(settings)
    settfile.close()


def execute():
    parser = argparse.ArgumentParser()
    parser.add_argument('spider_name', help='name of spider')
    args = parser.parse_args()

    generate(args.spider_name)

if __name__ == "__main__":
    execute()



import json
import re
import logging
from defaultsettings import CrawlSettings, RequestSettings, SpiderSettings
import redisds
from http import Request, RequestError


class Crawl:
    settings = CrawlSettings()

    def __init__(self, spider):
        redis_args = dict(host=self.settings.REDIS_URL,
                          port=self.settings.REDIS_PORT,
                          db=self.settings.REDIS_DB)
        self.url_set = redisds.Set('urlset', spider._name, **redis_args)
        self.url_queue = redisds.Queue('urlqueue', spider._name, json,
                                       **redis_args)
        self.running_count = redisds.Counter("count", namespace=spider._name,
                                             **redis_args)
        self.allowed_urls_regex = re.compile(spider._allowed_urls_regex)
        self.spider = spider
        self.start()

    def start(self):
        request = self.spider._start
        if self.settings.MODE in ["NORM", "RESUME"]:
            if self.settings.MODE == "NORM":
                self.url_queue.clear()
                self.url_set.clear()
            self.running_count.set(0)
        if request.callback is None:
            request.callback = "parse"
        self.insert(request)

    def clear(self, finished):
        if self.settings.MODE == "NORM" or finished:
            self.url_queue.clear()
            self.url_set.clear()
        if self.settings.MODE != "DISTRIBUTE":
            self.running_count.set(0)

    def count(self):
        return self.running_count.get()

    def inc_count(self):
        self.running_count.inc()

    def decr_count(self):
        self.running_count.decr()

    def insert(self, request, check=True):
        if not isinstance(request, Request):
            return
        reqhash = request.get_unique_id()
        if check:
            if not self.allowed_urls_regex.match(request.url):
                return
            elif reqhash in self.url_set:
                return
        self.url_set.add(reqhash)
        self.url_queue.put(request.__dict__)
        del request


class Crawler():

    @classmethod
    def load_spider(Crawler, spider_class, settings):
        Crawl.settings = CrawlSettings(settings.CRAWL)
        Request.settings = RequestSettings(settings.REQUEST)
        settings = SpiderSettings(settings.SPIDER)
        spider = spider_class(settings)
        spider.logger = logging.getLogger(spider._name)
        Crawler.crawl = Crawl(spider)

    def process_url(self):
        crawl = Crawler.crawl
        logger = logging.getLogger("dragline")
        request = Request(None)
        retry = 0
        while True:
            args = crawl.url_queue.get(timeout=2)
            if args:
                retry = 0
                request._set_state(args)
                logger.info("Processing %s", request)
                crawl.inc_count()
                try:
                    response = request.send()
                    try:
                        callback = getattr(crawl.spider, request.callback)
                        if request.meta:
                            requests = callback(response, request.meta)
                        else:
                            requests = callback(response)
                    except:
                        logging.exception("Failed to execute callback")
                        requests = None
                    if requests:
                        for i in requests:
                            crawl.insert(i)
                except RequestError as e:
                    request.retry += 1
                    if request.retry >= crawl.settings.MAX_RETRY:
                        logger.warning("Rejecting %s", request)
                    else:
                        logger.debug("Retrying %s", request)
                        crawl.insert(request, False)
                except KeyboardInterrupt:
                    pass
                except Exception as e:
                    logger.exception('Failed to open the url %s', request)
                else:
                    logger.info("Finished processing %s", request)
                finally:
                    crawl.decr_count()
            else:
                if crawl.count() == 0 or retry >= 2:
                    break
                else:
                    retry += 1
                logger.debug("Waiting for %s", crawl.count())



import redis


class Queue(object):

    """Simple Queue with Redis Backend"""

    def __init__(self, name, namespace='queue', serializer=None,
                 **redis_kwargs):
        """
        The default parameters are:
            namespace='queue', serializer=None, hash_func=usha1
            host='localhost', port=6379, db=0
        """
        self.__db = redis.Redis(**redis_kwargs)
        self.key = '%s:%s' % (namespace, name)
        self.serializer = serializer

    def __len__(self):
        return self.qsize()

    def clear(self):
        """Deletes a list"""
        self.__db.delete(self.key)

    def qsize(self):
        """Return the approximate size of the queue."""
        return self.__db.llen(self.key)

    def empty(self):
        """Return True if the queue is empty, False otherwise."""
        return self.qsize() == 0

    def put(self, item):
        """Put item into the queue."""
        if self.serializer:
            item = self.serializer.dumps(item)
        self.__db.rpush(self.key, item)

    def get(self, block=True, timeout=None):
        """Remove and return an item from the queue.

        If optional args block is true and timeout is None (the default), block
        if necessary until an item is available."""
        if block:
            item = self.__db.blpop(self.key, timeout=timeout)
        else:
            item = self.__db.lpop(self.key)
        if item:
            item = item[1]
            if self.serializer:
                item = self.serializer.loads(item)
        return item

    def get_nowait(self):
        """Equivalent to get(False)."""
        return self.get(False)


class Set(object):

    "A simple set with redis backend"

    def __init__(self, name, namespace='set', **redis_kwargs):
        """The default connection parameters are: host='localhost', port=6379, db=0"""
        self.__db = redis.Redis(**redis_kwargs)
        self.key = '%s:%s' % (namespace, name)

    def __len__(self):
        "returns the size of set"
        return self.__db.scard(self.key)

    def empty(self):
        "check whether the set is empty"
        return len(self) == 0

    def add(self, item):
        "Add an item to the set"
        self.__db.sadd(self.key, item)

    def __contains__(self, item):
        return self.is_member(item)

    def is_member(self, item):
        "Checks whtether a item is present in a set"
        return self.__db.sismember(self.key, item)

    def remove(self, item):
        "Remove an element from set"
        self.__db.srem(self.key, item)

    def clear(self):
        """Delete set"""
        self.__db.delete(self.key)


class Counter(object):
    def __init__(self, name, value=None, namespace='counter', **redis_kwargs):
        """The default connection parameters are: host='localhost', port=6379, db=0"""
        self.__db = redis.Redis(**redis_kwargs)
        self.key = '%s:%s' % (namespace, name)
        if value is not None:
            self.set(value)

    def inc(self, value=1):
        self.__db.incr(self.key, value)

    def decr(self, value=1):
        self.__db.decr(self.key, value)

    def set(self, value):
        self.__db.set(self.key, value)

    def get(self):
        value = self.__db.get(self.key)
        if value is None:
            return 0
        else:
            return int(value)



import os
import inspect
from httplib2 import Http
from urllib import urlencode
import base64
import subprocess
from os.path import normpath, basename
from runner import load_module
import argparse


def zipdir(path, zip):
    for root, dirs, files in os.walk(path):
        for file in files:
            print file
            zip.write(os.path.join(root, file))


def deploy(url, username, password, foldername, spider_website=None):
    # check whether the folder is a spider
    if not "main.py" in os.listdir(foldername):
        return "Not a valid spider"

    # check if the main.py contain a spider class
    module = load_module(foldername, "main")
    # check if main.py contain a spider class
    try:
        spider = getattr(module, "Spider")
    except Exception as e:
        return "Spider class not found"

    if not inspect.isclass(spider):
        return "Spider class not found"

    # create a spider object and check whether it contain required attributes
    spider_object = spider(None)

    try:
        if spider_object._name and spider_object._start and spider_object._allowed_urls_regex:
            spider_name = spider_object._name
        else:
            return "required attributes not found in spider"

    except Exception as e:
        print e
        return "Spider deploying failed"

    # zip the folder
    chfold = os.path.split(os.path.abspath(foldername))[0]
    os.chdir(chfold)
    comm = ["zip", "-r", "/tmp/%s.zip" %
            spider_name, ".", "-i", basename(normpath(foldername)) + "/*.py"]
    subprocess.call(comm)
    zipf = base64.encodestring(open("/tmp/%s.zip" % spider_name, "rb").read())
    post_data = {'username': username, 'password': password, 'name':
                 spider_name, 'zipfile': zipf}
    if spider_website:
        post_data['website'] = spider_website
    headers = {'content-type': 'application/x-www-form-urlencoded'}
    h = Http()
    resp, content = h.request(
        url, "POST", body=urlencode(post_data), headers=headers)
    # read zip file
    return content

def deploy_spider():
    parser = argparse.ArgumentParser()
    parser.add_argument('url')
    parser.add_argument('username')
    parser.add_argument('password')
    parser.add_argument('spider_dir')

    args = parser.parse_args()
    result = deploy(args.url, args.username, args.password, args.spider_dir)
    print result



if __name__ == "__main__":
    deploy_spider()




class Settings:
    def __init__(self, data={}):
        self.__dict__.update(data)

    LOGCONFIG = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'standard': {
                'format': '%(asctime)s [%(levelname)s] %(name)s: %(message)s'
            },
        },
        'handlers': {
            'default': {
                'level': 'DEBUG',
                'class': 'logging.StreamHandler',
                'formatter': 'standard'
            },
        },
        'loggers': {
            '': {
                'handlers': ['default'],
                'level': 'DEBUG',
                'propagate': True
            },
            'dragline': {
                'handlers': ['default'],
                'level': 'INFO',
                'propagate': False
            }
        }
    }


class RequestSettings(Settings):
    HEADERS = {
        "user-agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/33.0.1750.152 Safari/537.36",
        "accept": "text/html",
        'content-type': "application/x-www-form-urlencoded"
    }
    CACHE = None
    DELAY = 0.5
    MIN_DELAY = 0.5
    MAX_DELAY = 60


class CrawlSettings(Settings):
    MODE = "NORM"
    MAX_RETRY = 3
    REDIS_URL = 'localhost'
    REDIS_PORT = 6379
    REDIS_DB = 1


class SpiderSettings(Settings):
    pass
